{-# LANGUAGE RecordWildCards #-}
--------------------------------------------------------------------------------
-- |
-- Module      : TreasureHunters.Game
-- Copyright   : (c) Milan Straka, 2015
-- License     : BSD3
--
-- Maintainer  : fox@ucw.cz
-- Stability   : provisional
-- Portability :
--------------------------------------------------------------------------------

module TreasureHunters.Game ( bSize
                            , Board()
                            , showBoard
                            , showBoardForCurrent
                            , readBoard

                            , Game()
                            , Round()
                            , Move()

                            , newGame
                            , performRound
                            , isGameRunning
                            , gameWithOpponent
                            , showGame
                            , showGameForCurrent
                            , showGameHTML
                            , showGameHistoryHTML

                            , opponentTrivial
                            , opponentSimple
                            ) where

import Control.Applicative
import Data.Array
import Data.Char (isSpace, isDigit)
import Data.Function (on)
import Data.List (elemIndex, minimumBy, maximumBy)
import qualified Data.Map as Map
import Data.Maybe (fromJust, isNothing)
import qualified Data.Sequence as Seq
import Data.Time
import Data.Time.Clock.POSIX (getPOSIXTime)
import Data.Tuple (swap)
import System.Directory (createDirectory)
import System.FilePath ((</>))
import System.IO
import System.Locale (defaultTimeLocale)
import Text.Printf (printf)

import TreasureHunters.HTML
import TreasureHunters.Utils

---- Board ----
bSize :: Coord
bSize = 10

type Coord = Int {- 1..bSize -}
type Treasure = Int

data Field = FieldEmpty | FieldWall | FieldBase | FieldOpponentBase | FieldTreasure Treasure deriving Eq
data CurrentPlayer = Player1 | Player2 deriving Eq
data Player = Player { pcol :: Coord
                     , prow :: Coord
                     , carrying :: Treasure
                     , total :: Treasure
                     }
data Board = Board { fields :: Array (Coord, Coord) Field
                   , player1 :: Player
                   , player2 :: Maybe Player
                   , currentPlayer :: CurrentPlayer
                   , remainingMoves :: Int
                   }

isTreasure :: Field -> Bool
isTreasure (FieldTreasure _) = True
isTreasure _ = False

fieldString :: CurrentPlayer -> Field -> String
fieldString _ FieldEmpty = ".."
fieldString _ FieldWall = "XX"
fieldString Player1 FieldBase = "++"
fieldString Player2 FieldBase = "--"
fieldString Player1 FieldOpponentBase = "--"
fieldString Player2 FieldOpponentBase = "++"
fieldString _ (FieldTreasure treasure) = printf "%02d" treasure

showBoard :: Board -> ShowS
showBoard = showBoardFor Player1

showBoardForCurrent :: Board -> ShowS
showBoardForCurrent board@Board {..} = showBoardFor currentPlayer board

showBoardFor :: CurrentPlayer -> Board -> ShowS
showBoardFor current Board {..} = move . board
  where move = showString $ printf "%d\n" remainingMoves
        board = showString $ unlines $ map line [1..bSize]
        line row = concat [ fieldString current $ fields!(row,col) | col<-[1..bSize]]

showPlayersFor :: CurrentPlayer -> Board -> ShowS
showPlayersFor current Board {..} = player . otherPlayer
  where player | current == Player1 = showPlayer player1
               | current == Player2 = showPlayer $ fromJust player2
        otherPlayer | isNothing player2 = id
                    | current == Player1 = showPlayer (fromJust player2)
                    | current == Player2 = showPlayer player1
        showPlayer Player {..} = showString $ printf "%d %d %d %d\n" pcol prow carrying total

showBoardHTML :: Board -> ShowS
showBoardHTML Board {..} = monospaced (move . board)
  where move = showString $ printf "Zbývá %d tahů\n" remainingMoves
        board = showString $ unlines $ map line [1..bSize]
        line row = concat [ field (row,col) | col<-[1..bSize]]
        field pos@(row,col) | col == pcol player1 && row == prow player1 = showPlayer1 (carrying player1)
                            | Just p2 <- player2, col == pcol p2 && row == prow p2 = showPlayer2 (carrying p2)
                            | otherwise = case fields!pos of
                                FieldBase -> colorText "blue" $ fieldString Player1 FieldBase
                                FieldOpponentBase -> colorText "red" $ fieldString Player1 FieldOpponentBase
                                field -> fieldString Player1 field
        showPlayer1 0 = colorText "blue" "□□"
        showPlayer1 _ = colorText "blue" "■■"
        showPlayer2 0 = colorText "red" "○○"
        showPlayer2 _ = colorText "red" "●●"
        colorText color content =  printf "<span style='color:%s'>%s</span>" color content

readBoard :: String -> Either String Board
readBoard str = do
  (line, str) <- readLine str
  remainingMoves <- readWithFailure "Nelze načíst počet tahů" line
  (fields, str) <- readFields str
  verify "Na herní ploše chybí základna" $ sum (map (length . filter (==FieldBase)) fields) > 0
  verify "Na herní ploše je více základen" $ sum (map (length . filter (==FieldBase)) fields) <= 1
  verify "Na herní ploše je více soupeřových základen" $ sum (map (length . filter (==FieldOpponentBase)) fields) <= 1
  let Just player1 = createPlayer FieldBase fields
      player2 = createPlayer FieldOpponentBase fields
  return Board { fields=listArray ((1,1),(bSize,bSize)) (concat fields), currentPlayer=Player1, .. }
 where readLine str | (line, '\n':str) <- break (== '\n') str = return (removeCr line, str)
                    | otherwise = failure "Chybějící řádek při načítání herní plochy"
        where removeCr line | '\r' <- last line = init line
                            | otherwise = line
       readLines i str | i <= 0 = return ([], str)
                       | otherwise = do (line, str) <- readLine str
                                        (lines, str) <- readLines (i-1) str
                                        return (line : lines, str)
       readFields str = do (lines, str) <- readLines bSize str
                           fields <- mapM parseFields lines
                           return (fields, str)
       parseFields line = do let lineNoSpaces = filter (not . isSpace) line
                             verify (printf "Následující řádek herní plohy obsahuje špatný počet políček: '%s'" line) $ length lineNoSpaces == 2*bSize
                             mapM parseField $ splitFields lineNoSpaces
       parseField ".." = return FieldEmpty
       parseField "XX" = return FieldWall
       parseField "++" = return FieldBase
       parseField "--" = return FieldOpponentBase
       parseField treasure@[x,y] | isDigit x && isDigit y = FieldTreasure <$> readWithFailure (printf "Nemohu načíst hodnotu pokladu z políčka '%s'" treasure) treasure
       parseField field = failure $ printf "Neplatné políčko '%s'" field

       splitFields [] = []
       splitFields (x:y:zs) = [x,y] : splitFields zs

       createPlayer field fields = locate 1 fields
         where locate _ [] = Nothing
               locate r (row:rows) | Just c <- elemIndex field row = Just Player { prow=r, pcol=c+1, carrying=0, total=0 }
                                   | otherwise = locate (r+1) rows


---- Game ----
data Game = Game { board :: Either String Board
                 , rounds :: [Round]
                 , timeout :: Int
                 , logger :: Maybe Handle
                 }

data Round = Round { roundStart :: Clock
                   , roundBoard :: Board
                   , roundChosen :: Maybe Move
                   }

type Move = Char
type Clock = Integer

currentClock :: IO Clock
currentClock = do
  now <- getPOSIXTime
  return $ floor $ 1000 * toRational now

secToClock :: Int -> Clock
secToClock sec = 1000 * toInteger sec

---- Game logging ----
gameLogCreate :: Game    -- current game
              -> IO Game -- resulting game
gameLogCreate game = do
  try (createDirectory "logy") ()
  fname <- formatTime defaultTimeLocale ("logy" </> "%H-%M-%S%Q.txt") <$> (utcToLocalTime <$> getCurrentTimeZone <*> getCurrentTime)
  logger <- try (Just `fmap` openFile fname WriteMode) Nothing
  return game { logger=logger }

gameLog :: Game    -- current game
        -> String  -- string to log
        -> IO Game -- resulting game
gameLog game message = case logger game of
  Nothing -> return game
  Just logger -> try (hPutStr logger message >> hFlush logger >> return game) (game {logger=Nothing})

gameLogClose :: Game    -- current game
             -> IO Game -- resulting game
gameLogClose game = case logger game of
  Nothing -> return game
  Just logger -> try (hClose logger) () >> return game {logger=Nothing}

---- Game logic ----
newGame :: Board     -- starting board
        -> Int       -- timeout
        -> IO Game   -- resulting game
newGame board timeout = do
  let game = Game { board=return board, rounds=[], timeout=timeout, logger=Nothing }
  game <- gameLogCreate game
  game <- gameLog game $ showBoard board $ printf "Timeout: %d\n" timeout
  nextRound game =<< currentClock

nextRound :: Game     -- game with finished round
          -> Clock    -- current time
          -> IO Game  -- new game with open round
nextRound game@Game{board = Left _} _ = return game
nextRound game@Game{board = Right board, ..} now
  | remainingMoves board <= 0 = return game { board = failure endOfGameMessage, rounds = round : rounds }
  | otherwise = do game <- gameLog game "Zvoleny pohyb: "
                   return game { rounds = round : rounds }
 where round = Round { roundStart=now, roundBoard=board, roundChosen=Nothing}
       endOfGameMessage | Just p2 <- player2 board = printf "Konec hry: %d %d" (total (player1 board)) (total p2)
                        | otherwise = printf "Konec hry: %d" $ total (player1 board)

showGame :: Int -> Game -> ShowS
showGame = showGameFor Player1

showGameForCurrent :: Int -> Game -> ShowS
showGameForCurrent gid game@Game { board = Right Board {..} } = showGameFor currentPlayer gid game
showGameForCurrent gid game = showGameFor Player1 gid game

showGameFor :: CurrentPlayer -> Int -> Game -> ShowS
showGameFor current gid game@Game { board = Left message } = showString message
showGameFor current gid game@Game { board = Right board } =
  shows gid . showChar '\n' . showBoardFor current board . showPlayersFor current board

showGameHTML :: Game -> ShowS
showGameHTML game@Game { board = Left message } = showString "Hra skončila zprávou:<br>" . showString message
showGameHTML game@Game { board = Right board } =
  showString "Hra probíhá:\n" . showBoardHTML board

showGameHistoryHTML :: Game -> ShowS
showGameHistoryHTML game = table $ showRounds 1 (reverse $ rounds game)
 where
  showRounds i [r] = [showRound i r Nothing (either id (const "Hra stále probíhá") (board game))]
  showRounds i (r:r':rs) = showRound i r (Just $ roundStart r') "" : showRounds (i+1) (r':rs)
  showRound i Round {..} nextRoundStart remark = [ (TableAlignCenter, showString $ "Kolo " ++ show i)
                                                 , (TableAlignCenter, showBoardHTML roundBoard)
                                                 , (TableAlignLeft, showString statusStr) ]
   where
    chosenStr | isNothing roundChosen = ""
              | Just chosen <- roundChosen, isNothing nextRoundStart = printf "%c" chosen
              | Just chosen <- roundChosen, Just nextRoundStart <- nextRoundStart =
                  printf "%c po %dms" chosen (nextRoundStart - roundStart)
    statusStr = chosenStr ++ (if null chosenStr || null remark then "" else "<br>") ++ remark

isGameRunning :: Game -> Bool
isGameRunning Game { board = Right board } = True
isGameRunning Game { board = Left message } = False

gameWithOpponent :: Game -> Bool
gameWithOpponent Game { board = Right Board { player2 = Just _ } } = True
gameWithOpponent _ = False

performRound :: Game    -- unfinished game
             -> Move    -- moves for this round
             -> IO Game -- resulting game
performRound game@Game { board=Left _ } _ = return game
performRound game@Game { board=Right board } move = do
  now <- currentClock
  game <- gameLog game { rounds=store move game } $ printf "%c po %dms\n" move (now - timeOf game)
  case game of
   Game { timeout=timeout, rounds=Round {..} : _ }
    | timeout > 0 && now > roundStart + secToClock timeout ->
        endGame game $ printf "Chyba: Tah trval moc dlouho (%dms > %ds)." (now - roundStart) timeout
    | otherwise -> case performMove move board of
                     Left message -> endGame game message
                     Right board -> nextRound game { board=Right board} now

 where
  store move Game { rounds=round:rounds } = round { roundChosen = Just move } : rounds
  timeOf Game { rounds=Round {roundStart=roundStart}:_ } = roundStart

  endGame :: Game -> String -> IO Game
  endGame game message = do
    game <- gameLog game (message ++ "\n")
    game <- gameLogClose game
    return game { board=Left message }

  performMove :: Move -> Board -> Either String Board
  performMove move Board {..} | Just p2 <- player2, currentPlayer == Player2 = do
    (nextP2, nextFields) <- perform move fields p2 FieldOpponentBase FieldBase (Just player1)
    return Board { fields=nextFields, player1=player1, player2=Just nextP2, currentPlayer=Player1, remainingMoves=remainingMoves-1 }
                              | otherwise = do
    (nextP1, nextFields) <- perform move fields player1 FieldBase FieldOpponentBase player2
    return Board { fields=nextFields, player1=nextP1, player2=player2, currentPlayer=if isNothing player2 then Player1 else Player2, remainingMoves=remainingMoves-1 }

   where perform move fields player@Player {..} playerBase opponentBase opponent = do
           (cDelta, rDelta) <- parseMove move
           let (newCol, newRow) | validPosition (pcol+cDelta) (prow+rDelta) = (pcol + cDelta, prow + rDelta)
                                | otherwise = (pcol, prow)
           return $ handleTreasures player { pcol=newCol, prow=newRow }
          where parseMove 'L' = return (-1, 0)
                parseMove 'P' = return ( 1, 0)
                parseMove 'N' = return (0, -1)
                parseMove 'D' = return (0,  1)
                parseMove move = failure $ printf "Neznamy pohyb '%c'" move

                validPosition c r | c < 1 || c > bSize || r < 1 || r > bSize = False
                                  | fields!(r,c) == FieldWall = False
                                  | fields!(r,c) == opponentBase = False
                                  | Just Player {..} <- opponent, r == prow && c == pcol = False
                                  | otherwise = True
                handleTreasures player@Player {..} | FieldTreasure treasure <- fields!(prow,pcol), carrying == 0 =
                                                       (player { carrying=treasure}, fields // [((prow,pcol), FieldEmpty)])
                                                   | fields!(prow,pcol) == playerBase =
                                                       (player { carrying=0, total=total+carrying}, fields)
                                                   | otherwise = (player, fields)

opponentTrivial :: Game -> Move
opponentTrivial game@Game {board=Right Board { player2=Just player2, ..}} = opponentUsingPolicy game trivialPolicy
 where trivialPolicy playerPos basePos distancesFromOpponent distancesFromBase
         | carrying player2 > 0 = moveTo distancesFromBase playerPos basePos
         | otherwise = case filter (isTreasure . dField . snd) $ Map.assocs distancesFromOpponent of
                         [] -> 'N'
                         treasures -> moveTo distancesFromOpponent playerPos $ fst $ minimumBy (compare `on` (dDistance . snd)) treasures

opponentSimple :: Game -> Move
opponentSimple game@Game {board=Right Board { player2=Just player2, ..}} = opponentUsingPolicy game simplePolicy
 where simplePolicy playerPos basePos distancesFromOpponent distancesFromBase
         | carrying player2 > 0 = moveTo distancesFromBase playerPos basePos
         | otherwise = let distancesFromBoth = Map.intersectionWith (\d e -> (case dField d of FieldTreasure treasure -> treasure, dDistance d + dDistance e)) distancesFromBase $ Map.filter (isTreasure . dField) distancesFromOpponent
                       in case filter (reachableTreasure . snd . snd) $ Map.assocs distancesFromBoth of
                            [] -> 'N'
                            treasures -> moveTo distancesFromOpponent playerPos $ fst $ maximumBy (\(_,(t1,l1)) (_,(t2,l2)) -> compare (t1*l2) (t2*l1)) $ treasures
       reachableTreasure steps | steps <= (remainingMoves + 1) `div` 2 = True
       reachableTreasure _ = False

opponentUsingPolicy :: Game -> ((Coord, Coord) -> (Coord, Coord) -> DistanceInfos -> DistanceInfos -> Move) -> Move
opponentUsingPolicy Game { board=Right Board { currentPlayer = Player2, player2 = Just player2, ..} } policy =
  let player1Pos = (prow player1, pcol player1)
      player2Pos = (prow player2, pcol player2)
      basePos = fromJust $ lookup FieldOpponentBase $ map swap $ assocs fields
      fieldsWithPlayer = fields // [((prow player1, pcol player1), FieldWall)]
      distancesFromOpponent = computeDistances fieldsWithPlayer (prow player2, pcol player2) False
      distancesFromBase = computeDistances fieldsWithPlayer basePos True
  in policy player2Pos basePos distancesFromOpponent distancesFromBase

data DistanceInfo = DistanceInfo { dField :: Field, dDistance :: Int, dMoves :: [Move] }
type DistanceInfos = Map.Map (Coord, Coord) DistanceInfo

computeDistances :: Array (Coord, Coord) Field -> (Coord, Coord) -> Bool -> DistanceInfos
computeDistances fields start throughTreasures = bfs (Seq.singleton start) $ Map.singleton start DistanceInfo { dField=fields!start, dDistance=0, dMoves=[] }
 where bfs queue map = case Seq.viewl queue of
         Seq.EmptyL -> map
         (Seq.:<) q queue | Just qDistanceInfo <- q `Map.lookup` map ->
           let (newQueue, newMap) | not throughTreasures && isTreasure (dField qDistanceInfo) = (queue, map)
                                  | otherwise = probe (neighbors q) qDistanceInfo queue map
           in bfs newQueue newMap
       neighbors (r, c) = [ ('N', (r-1, c)), ('D', (r+1, c)), ('L', (r, c-1)), ('P', (r, c+1)) ]
       probe [] _ queue map = (queue, map)
       probe ((move,p@(r,c)):fs) qDistanceInfo queue map
         | r < 1 || r > bSize || c < 1 || c > bSize ||
           fields!p == FieldWall || fields!p == FieldBase || Map.member p map
           = probe fs qDistanceInfo queue map
         | otherwise = probe fs qDistanceInfo ((Seq.|>) queue p) (Map.insert p DistanceInfo
            { dField=fields!p, dDistance=1+dDistance qDistanceInfo, dMoves=move:dMoves qDistanceInfo} map)

moveTo :: DistanceInfos -> (Coord, Coord) -> (Coord, Coord) -> Move
moveTo distances from to | Just dFrom <- from `Map.lookup` distances, dDistance dFrom == 0, Just dTo <- to `Map.lookup` distances = last $ dMoves dTo
                         | Just dFrom <- from `Map.lookup` distances = reverseMove $ head $ dMoves dFrom
                         | otherwise = 'N'
 where reverseMove 'L' = 'P'
       reverseMove 'P' = 'L'
       reverseMove 'N' = 'D'
       reverseMove 'D' = 'N'
