--------------------------------------------------------------------------------
-- |
-- Module      : TreasureHunters.Server.HTTPServer
-- Copyright   : (c) Milan Straka, 2015
-- License     : BSD3
--
-- Maintainer  : fox@ucw.cz
-- Stability   : provisional
-- Portability :
--------------------------------------------------------------------------------

module TreasureHunters.Server.HTTPServer ( Server
                                         , startServer

                                         , ServerResponse
                                         , plainResponse
                                         , htmlResponse
                                         , mimeResponse
                                         ) where

import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Char (ord)
import Data.Time (getCurrentTime, formatTime)
import Network.Socket
import Network.URL
import System.IO
import System.Locale (defaultTimeLocale)
import Text.Printf

import TreasureHunters.Server.Resources
import TreasureHunters.Utils

#ifdef _OS_UNIX
import qualified System.Posix.Signals as P
                        (installHandler,sigPIPE,Handler(Ignore))
#endif


type Server = String                 -- path
           -> [(String, String)]     -- params
           -> MVar ()                -- exit flag
           -> IO ServerResponse      -- response

type ServerResponse = Maybe ({-Mime-} String, {-Body-} String)

plainResponse :: ShowS -> ServerResponse
plainResponse text = Just ("text/plain; charset=utf-8", text "\n")

htmlResponse :: ShowS -> ServerResponse
htmlResponse text = Just ("text/html; charset=utf-8", text "\n")

mimeResponse :: String -> ShowS -> ServerResponse
mimeResponse mime text = Just (mime, text "\n")

startServer :: PortNumber -- port number
            -> Server     -- server
            -> IO ()
startServer port server = withSocketsDo $ do
#ifdef _OS_UNIX
  P.installHandler P.sigPIPE P.Ignore Nothing
#endif
  endVar <- newEmptyMVar
  sock <- server_socket
  forkIO $ forever (loop sock endVar) `finally` putMVar endVar ()
  takeMVar endVar

 where
  server_socket = do
    sock <- socket AF_INET Stream defaultProtocol
#ifdef _OS_UNIX
    setSocketOption sock ReuseAddr 1
#endif
    bindSocket sock (SockAddrInet port iNADDR_ANY)
    listen sock maxListenQueue
    return sock

  loop sock endVar = do
    (client, _) <- accept sock
    client <- socketToHandle client ReadWriteMode
    hSetEncoding client utf8
    hSetNewlineMode client $ NewlineMode { inputNL = CRLF, outputNL = CRLF }
    hSetBuffering client LineBuffering
    forkIO (handle (showException "Pri zpracovani pozadavku doslo k chybe (ale server bude dal bezet)") (serve client endVar) >> hClose client)

  serve client endVar = unlessEOF client $ do
    request <- hGetLine client
    case words request of
      method' : url' : version' | Just version <- parseHTTPVersion version'
                                , Just url <- importURL url'
                                -> process_request method' url version
      _ -> respond 11 "400 Bad Request" $ plain $ printf "Bad request %s." request

    where
      process_request method url version
       | method == "GET" = do when (version > 9) skip_headers
                              response <- server_with_resources (url_path url) (url_params url) endVar
                              case response of
                                Just mimeBody -> respond version "200 OK" mimeBody
                                Nothing -> respond version "404 Not Found" $ plain $ printf "Document %s not found." (exportURL url)
       | otherwise = respond version "501 Not Implemented" $ plain $ printf "Method %s not implemented." method

      skip_headers = unlessEOF client $ do line <- hGetLine client
                                           unless (line == "") skip_headers

      server_with_resources "bootstrap.min.css" _ _ = return $ mimeResponse "text/css" $ showString bootstrapCSS
      server_with_resources "bootstrap.min.js" _ _ = return $ mimeResponse "text/javascript" $ showString bootstrapJS
      server_with_resources "jquery.min.js" _ _ = return $ mimeResponse "text/javascript" $ showString jQueryJS
      server_with_resources url params endVar = server url params endVar

      respond version code (mime, body) = do
        headers <- responseHeaders
        hPutStr client $ unlines headers ++ body

       where
        responseHeaders
         | version == 9 = return []
         | version == 10 || version == 11 = do
           date <- currentHTTPDate
           let http10headers = [ "HTTP/1." ++ show (version `mod` 10) ++ " " ++ code
                               , "Date: " ++ date
                               , "Server: Treasure Hunters server"
                               , "Content-Type: " ++ mime
                               , "Content-Length: " ++ show (encodedLength body 0)
                               , "Pragma: no-cache"
                               , "Access-Control-Allow-Origin: *" ]
               http11headers = [ "Cache-Control: no-cache, max-age=0, must-revalidate"
                               , "Connection: close" ]
           return $ http10headers ++ (if version == 11 then http11headers else []) ++ [""]

        currentHTTPDate = formatTime defaultTimeLocale "%a, %d %b %Y %H:%M:%S GMT" `fmap` getCurrentTime

        encodedLength [] acc = acc :: Int
        encodedLength (c:cs) acc = encodedLength cs $ acc + case ord c of o | c == '\n' -> 2
                                                                            | o <= 0x7f -> 1
                                                                            | o <= 0x7ff -> 2
                                                                            | o <= 0xffff -> 3
                                                                            | otherwise -> 4


      plain str = ("text/plain; charset=utf-8", str ++ "\n")

      parseHTTPVersion :: [String] -> Maybe Int
      parseHTTPVersion [] = Just 9
      parseHTTPVersion ["HTTP/1.0"] = Just 10
      parseHTTPVersion ["HTTP/1.1"] = Just 11
      parseHTTPVersion _ = Nothing

      unlessEOF handle action = do
        eof <- hIsEOF handle
        unless eof action
