{-# LANGUAGE RecordWildCards #-}
--------------------------------------------------------------------------------
-- |
-- Module      : TreasureHunters.Server.Server
-- Copyright   : (c) Milan Straka, 2015
-- License     : BSD3
--
-- Maintainer  : fox@ucw.cz
-- Stability   : provisional
-- Portability :
--------------------------------------------------------------------------------

module TreasureHunters.Server.Server ( initServer ) where

import Control.Concurrent
import Control.Exception (handle, SomeException)
import Control.Monad
import Data.Char (toUpper)
import qualified Data.IntMap as M
import Data.List (sort)
import Network.URL
import System.Directory
import System.FilePath
import Text.Printf

import TreasureHunters.Game
import TreasureHunters.HTML
import TreasureHunters.Server.HTTPServer
import TreasureHunters.Server.Resources
import TreasureHunters.Utils

defaultBoard :: Board
defaultBoard = board
  where (Right board) = readBoard defaultMap

data Opponent = OpponentTrivial | OpponentSimple | OpponentClient (Chan ()) deriving Eq
type GameWithOpponent = (Game, Maybe Opponent)

data State = State { newBoard :: Board
                   , newOpponent :: Opponent
                   , newTimeout :: Int
                   , games :: M.IntMap (MVar GameWithOpponent)
                   , startingGame :: Maybe Int
                   }

initServer :: Maybe String -- initial map to load
           -> Int          -- initial timeout
           -> IO Server    -- created server
initServer map timeout = do
  board <- maybe (return defaultBoard) ((either error return . readBoard =<<) . readFile . ("mapy" </>)) map
  state <- newMVar State { newBoard=board, newOpponent=OpponentTrivial, newTimeout=timeout, games=M.empty, startingGame=Nothing }
  return (server state)

server :: MVar State -- global server state
       -> Server     -- URL handler
server stateVar str params endVar = case str of
  "" -> mainPage
  "hra" -> needArgs ["id"] gamePage
  "nastav" -> setParams params
  "reset" -> resetGames >> return (htmlResponse $ htmlRedirect "")
  "konec" -> forkIO (threadDelay (10^6) >> putMVar endVar ()) >> return (htmlResponse $ htmlPage $ heading "Nashledanou" id)

  "start" -> startGame
  "tah" -> needArgs ["id", "pohyb"] gameRound

  "jsklient" -> return $ htmlResponse $ jsHtmlPage $ showString clientHTML

  _ -> return Nothing

 where
  mainPage = do
    newGamePage <- newGamePage
    gamesPage <- gamesPage
    return $ htmlResponse $ htmlPage $ heading "Lovci pokladů" $ group "Server" serverPage . group "Nová hra" newGamePage . group "Probíhající hry" gamesPage
   where
    serverPage = table [[ (TableAlignCenter, form "reset" $ button "Reset serveru")
                        , (TableAlignCenter, form "konec" $ button "Zastavit server") ]]
    newGamePage = do
      state <- readMVar stateVar
      maps <- filterM (doesFileExist . ("mapy" </>)) =<< try (getDirectoryContents "mapy") []
      return $ table [[ (TableAlignCenter, newGamePane state), (TableAlignCenter, loadMapPane maps) ]]
     where
      newGamePane state = form "nastav" $
        table [ [ (TableAlignRight, showString "Mapa:"), (TableAlignLeft, textArea "mapa" (2*bSize) (bSize+1) $ showBoard (newBoard state)) ]
              , [ (TableAlignRight, showString "Timeout:"), (TableAlignLeft, input "timeout" $ show (newTimeout state)) ]
              , [ (TableAlignRight, showString "Protihráč:"), (TableAlignLeft, opponentRadios (newOpponent state)) ]
              , [ (TableAlignRight, showString ""), (TableAlignLeft, button "Nastavit") ] ]
      opponentRadios opponent = radio "protihrac" "trivial" (opponent == OpponentTrivial) "Triviální" .
                                radio "protihrac" "simple" (opponent == OpponentSimple) "Jednoduchý" .
                                radio "protihrac" "client" (case opponent of OpponentClient _ -> True; _ -> False) "Jiný hráč"
      loadMapPane maps = group "Existující mapy" $ fileBoards (sort maps)
      fileBoards [] = id
      fileBoards (map:maps) = href ("nastav?soubor=" ++ encString True ok_param map) map . showString "<br>" . fileBoards maps

    gamesPage = do
      State{..} <- readMVar stateVar
      table `fmap` mapM showGame (reverse $ M.assocs games)
     where showGame (id, gameOpponentVar) =
             do (game, _) <- readMVar gameOpponentVar
                return [ (TableAlignLeft, showString (printf "Hra číslo %d<br>" id) . href ("hra?id=" ++ show id) "Průběh hry")
                       , (TableAlignLeft, showGameHTML game) ]

  gamePage [id'] = do
    State{..} <- readMVar stateVar
    case maybeRead id' of
      Just id | Just gameOpponentVar <- M.lookup id games ->
        do (game, _) <- readMVar gameOpponentVar
           return $ htmlResponse $ htmlPage $ heading (printf "Hra číslo %d" id) $ showGameHistoryHTML game
      _ -> return $ plainResponse $ showString "Chyba: Nezname identifikacni cislo hry."

  setParams [] = return $ htmlResponse $ htmlRedirect ""
  setParams ((param, value) : params) =
    setParam param value >>= either (return . plainResponse . showString) (\stateMod -> do
    state <- takeMVar stateVar
    putMVar stateVar (stateMod state)
    setParams params)
   where
    setParam "mapa" value = return (readBoard value >>= \board -> return $ \state -> state { newBoard=board })
    setParam "soubor" value =
      handle (\e -> return $ failure $ "Chyba: " ++ show (e :: SomeException)) $ readFile ("mapy" </> value) >>= setParam "mapa"
    setParam "timeout" value | Just timeout <- maybeRead value = return $ return $ \state -> state { newTimeout=timeout }
                             | otherwise = return $ failure "Chyba: Timeout je ve spatnem formatu."
    setParam "protihrac" value | value == "trivial" = return $ return $ \state -> state { newOpponent=OpponentTrivial }
                               | value == "simple" = return $ return $ \state -> state { newOpponent=OpponentSimple }
                               | value == "client" = do chan <- newChan
                                                        return $ return $ \state -> state { newOpponent=OpponentClient chan }
                               | otherwise = return $ failure $ printf "Chyba: Neznamy popis protihrace %s." value
    setParam param value = return $ failure $ printf "Chyba: Neznamy parametr %s k nastaveni." param

  resetGames = do
    state <- takeMVar stateVar
    putMVar stateVar state { games = M.empty }

  startGame = do
    state@State{..} <- takeMVar stateVar
    case newOpponent of
      OpponentClient _ | Just gameId <- startingGame -> do
        let Just gameOpponentVar = M.lookup gameId games
        putMVar stateVar state { startingGame=Nothing }
        (_, Just (OpponentClient chan)) <- readMVar gameOpponentVar
        writeChan chan ()
        readChan chan
        (game, _) <- readMVar gameOpponentVar
        return $ plainResponse $ showGameForCurrent gameId game
      _ -> do
        game <- newGame newBoard newTimeout
        let opponent | gameWithOpponent game = Just newOpponent
                     | otherwise = Nothing
        gameOpponentVar <- newMVar (game, opponent)
        let gameId = 1 + if M.null games then 0 else fst (M.findMax games)
        case opponent of
          Just (OpponentClient chan) -> do
            newOpponentChan <- newChan
            putMVar stateVar state { games = M.insert gameId gameOpponentVar games
                                   , newOpponent=OpponentClient newOpponentChan
                                   , startingGame = Just gameId }
            readChan chan
            return $ plainResponse $ showGameForCurrent gameId game
          _ -> do
            putMVar stateVar state { games = M.insert gameId gameOpponentVar games }
            return $ plainResponse $ showGameForCurrent gameId game

  gameRound [id', move] = do
    State{..} <- readMVar stateVar
    case maybeRead id' of
      Just id | Just gameOpponentVar <- M.lookup id games ->
        do (game, opponent) <- takeMVar gameOpponentVar
           case isGameRunning game of
             True | length move == 0 -> return $ plainResponse $ showString "Chyba: Nebyl zadan zadny pohyb."
             True | length move  > 1 -> return $ plainResponse $ showString "Chyba: Byl zadan vic nez jeden pohyb."
             True -> do game <- performRound game $ toUpper $ head move
                        game <- case opponent of Just (OpponentClient chan) | not $ isGameRunning game -> writeChan chan () >> return game
                                                 _ | not $ isGameRunning game -> return game
                                                 Nothing -> return game
                                                 Just OpponentTrivial -> performRound game (opponentTrivial game)
                                                 Just OpponentSimple -> performRound game (opponentSimple game)
                                                 Just (OpponentClient chan) -> do
                                                   putMVar gameOpponentVar (game, opponent)
                                                   writeChan chan ()
                                                   readChan chan
                                                   fst `fmap` takeMVar gameOpponentVar
                        putMVar gameOpponentVar (game, opponent)
                        return $ plainResponse $ showGameForCurrent id game
             False -> do putMVar gameOpponentVar (game, opponent)
                         return $ plainResponse $ showString "Chyba: Hra s timto identifikacnim cislem uz skoncila."
      _ -> return $ plainResponse $ showString "Chyba: Nezname identifikacni cislo hry."

  needArgs names fun = needArgs' names []
    where needArgs' [] vals = fun (reverse vals)
          needArgs' (name:names) vals | Just val <- lookup name params = needArgs' names (val : vals)
                                      | otherwise = return $ plainResponse $ showString $ printf "Chyba: Chybi HTTP argument %s." name
